from datetime import datetime
from typing import Dict, List, Tuple, Union

import pandas as pd
from loguru import logger

from edsteva.probes.utils.prepare_df import (
    prepare_biology_relationship,
    prepare_care_site,
    prepare_measurement,
    prepare_person,
    prepare_visit_occurrence,
)
from edsteva.probes.utils.utils import (
    CARE_SITE_LEVEL_NAMES,
    concatenate_predictor_by_level,
    hospital_only,
    impute_missing_dates,
)
from edsteva.utils.checks import check_tables
from edsteva.utils.framework import is_koalas, to
from edsteva.utils.typing import Data, DataFrame


def compute_completeness_predictor_per_visit(
    self,
    data: Data,
    care_site_relationship: pd.DataFrame,
    start_date: datetime,
    end_date: datetime,
    care_site_levels: Union[bool, str, List[str]],
    stay_types: Union[bool, str, Dict[str, str]],
    care_site_ids: List[int],
    care_site_short_names: List[str],
    care_site_specialties: Union[bool, List[str]],
    concept_codes: Union[bool, List[str]],
    care_sites_sets: Union[str, Dict[str, str]],
    specialties_sets: Union[str, Dict[str, str]],
    concepts_sets: Union[str, Dict[str, str]],
    length_of_stays: List[float],
    source_terminologies: Dict[str, str],
    mapping: List[Tuple[str, str, str]],
    age_ranges: List[int],
    provenance_sources: Union[bool, str, Dict[str, str]],
    stay_sources: Union[bool, str, Dict[str, str]],
    **kwargs
):
    r"""Script to be used by [``compute()``][edsteva.probes.base.BaseProbe.compute]

    The ``per_visit`` algorithm computes $c_(t)$ the availability of laboratory data related linked to patients' administrative stays:

    $$
    c(t) = \frac{n_{with\,biology}(t)}{n_{visit}(t)}
    $$

    Where $n_{visit}(t)$ is the number of administrative stays, $n_{with\,condition}$ the number of stays having at least one biological measurement recorded and $t$ is the month.
    """
    self._metrics = ["c", "n_visit", "n_visit_with_measurement"]
    check_tables(
        data=data,
        required_tables=[
            "measurement",
            "concept",
            "concept_relationship",
            "visit_occurrence",
        ],
    )
    standard_terminologies = self._standard_terminologies
    biology_relationship = prepare_biology_relationship(
        data=data,
        standard_terminologies=standard_terminologies,
        source_terminologies=source_terminologies,
        mapping=mapping,
    )

    self.biology_relationship = biology_relationship
    root_terminology = mapping[0][0]

    person = prepare_person(data)

    visit_occurrence = prepare_visit_occurrence(
        data=data,
        start_date=start_date,
        end_date=end_date,
        stay_types=stay_types,
        length_of_stays=length_of_stays,
        provenance_sources=provenance_sources,
        stay_sources=stay_sources,
        person=person,
        age_ranges=age_ranges,
    )

    measurement = prepare_measurement(
        data=data,
        biology_relationship=biology_relationship,
        concept_codes=concept_codes,
        concepts_sets=concepts_sets,
        root_terminology=root_terminology,
        standard_terminologies=standard_terminologies,
        per_visit=True,
    )

    care_site = prepare_care_site(
        data=data,
        care_site_ids=care_site_ids,
        care_site_short_names=care_site_short_names,
        care_site_specialties=care_site_specialties,
        care_sites_sets=care_sites_sets,
        specialties_sets=specialties_sets,
        care_site_relationship=care_site_relationship,
    )

    hospital_visit = get_hospital_visit(
        self,
        measurement=measurement,
        visit_occurrence=visit_occurrence,
        care_site=care_site,
    )
    hospital_name = CARE_SITE_LEVEL_NAMES["Hospital"]
    biology_predictor_by_level = {hospital_name: hospital_visit}

    if care_site_levels and not hospital_only(care_site_levels=care_site_levels):
        logger.info(
            "Biological measurements are only available at hospital level for now"
        )
        care_site_levels = "Hospital"

    biology_predictor = concatenate_predictor_by_level(
        predictor_by_level=biology_predictor_by_level,
        care_site_levels=care_site_levels,
    )

    return compute_completeness(self, biology_predictor)


def compute_completeness(
    self,
    biology_predictor: DataFrame,
):
    # Visit with measurement
    partition_cols = [*self._index.copy(), "date"]
    n_visit_with_measurement = (
        biology_predictor.groupby(
            partition_cols,
            as_index=False,
            dropna=False,
        )
        .agg({"has_measurement": "count"})
        .rename(columns={"has_measurement": "n_visit_with_measurement"})
    )
    n_visit_with_measurement = to("pandas", n_visit_with_measurement)
    n_visit_with_measurement = n_visit_with_measurement[
        n_visit_with_measurement.n_visit_with_measurement > 0
    ]
    n_visit_with_measurement = impute_missing_dates(
        start_date=self.start_date,
        end_date=self.end_date,
        predictor=n_visit_with_measurement,
        partition_cols=partition_cols,
    )
    # Visit total
    biology_columns = ["concepts_set"] + [
        "{}_concept_code".format(terminology)
        for terminology in self._standard_terminologies
    ]
    partition_cols = list(set(partition_cols) - set(biology_columns))
    n_visit = (
        biology_predictor.groupby(
            partition_cols,
            as_index=False,
            dropna=False,
        )
        .agg({"visit_id": "nunique"})
        .rename(columns={"visit_id": "n_visit"})
    )
    n_visit = to("pandas", n_visit)
    n_visit = impute_missing_dates(
        start_date=self.start_date,
        end_date=self.end_date,
        predictor=n_visit,
        partition_cols=partition_cols,
    )

    biology_predictor = n_visit_with_measurement.merge(
        n_visit,
        on=partition_cols,
    )

    biology_predictor["c"] = biology_predictor["n_visit"].where(
        biology_predictor["n_visit"] == 0,
        biology_predictor["n_visit_with_measurement"] / biology_predictor["n_visit"],
    )

    return biology_predictor


def get_hospital_visit(
    self,
    measurement: DataFrame,
    visit_occurrence: DataFrame,
    care_site: DataFrame,
):
    hospital_measurement = measurement[
        set(measurement.columns).intersection(
            set(["visit_occurrence_id", *self._index])
        )
    ].drop_duplicates()
    hospital_measurement["has_measurement"] = True
    hospital_visit = visit_occurrence.merge(
        hospital_measurement,
        on="visit_occurrence_id",
        how="left",
    )
    hospital_visit = hospital_visit.rename(columns={"visit_occurrence_id": "visit_id"})
    hospital_visit = hospital_visit.merge(care_site, on="care_site_id")

    if is_koalas(hospital_visit):
        hospital_visit = hospital_visit.spark.cache()

    return hospital_visit



from copy import deepcopy
from datetime import datetime
from typing import Dict, List, Union

import altair as alt
import pandas as pd

from edsteva.models.base import BaseModel
from edsteva.probes.base import BaseProbe
from edsteva.viz.utils import (
    add_estimates_filters,
    configure_style,
    create_groupby_selection,
    filter_data,
    generate_error_line,
    generate_main_chart,
    generate_model_line,
    generate_probe_line,
    get_indexes_to_groupby,
    month_diff,
    save_html,
)


def normalized_probe_plot(
    probe: BaseProbe,
    fitted_model: BaseModel,
    care_site_level: List[str] = None,
    start_date: Union[datetime, str] = None,
    end_date: Union[datetime, str] = None,
    t_min: int = None,
    t_max: int = None,
    save_path: str = None,
    x_axis_title: str = None,
    y_axis_title: str = None,
    main_chart_config: Dict[str, float] = None,
    model_line_config: Dict[str, str] = None,
    probe_line_config: Dict[str, str] = None,
    error_line_config: Dict[str, str] = None,
    estimates_selections: Dict[str, str] = None,
    estimates_filters: Dict[str, str] = None,
    chart_style: Dict[str, float] = None,
    indexes_to_remove: List[str] = ["care_site_id"],
    **kwargs,
):
    r"""Displays a chart with the aggregated normalized completeness predictor $\frac{c(\Delta t)}{c_0}$ over normalized time $\Delta t = t - t_0$. It represents the overall deviation from the Model.

    Is is possible to save the chart in HTML with the "save_path" optional input.

    Parameters
    ----------
    probe : BaseProbe
        Class describing the completeness predictor $c(t)$
    fitted_model : BaseModel
        Model fitted to the probe
    care_site_level : List[str], optional
        **EXAMPLE**: `["Hospital"]`, `["Hôpital", "UF"]` or `["UF", "UH"]`
    start_date : datetime, optional
        **EXAMPLE**: `"2019-05-01"`
    end_date : datetime, optional
        **EXAMPLE**: `"2021-07-01"`
    t_min : int, optional
        Minimal difference with $t_0$ in month $\Delta t_{min}$.

        **EXAMPLE**: `-24`
    t_max : int, optional
        Maximal difference with $t_0$ in month $\Delta t_{max}$.

        **EXAMPLE**: `24`
    save_path : str, optional
        Folder path where to save the chart in HTML format.
    x_axis_title: str, optional,
        Label name for the x axis.
    y_axis_title: str, optional,
        Label name for the y axis.
    main_chart_config: Dict[str, str], optional
        If not None, configuration used to construct the top main chart.
    model_line_config: Dict[str, str], optional
        If not None, configuration used to construct the model line.
    error_line_config: Dict[str, str], optional
        If not None, configuration used to construct the error line.
    probe_line_config: Dict[str, str], optional
        If not None, configuration used to construct the probe line.
    estimates_selections: Dict[str, str], optional
        If not None, configuration used to construct the estimates selections.
    estimates_filters: Dict[str, str], optional
        If not None, configuration used to construct the estimates filters.
    chart_style: Dict[str, float], optional
        If not None, configuration used to configure the chart style.

        **EXAMPLE**: `{"labelFontSize": 13, "titleFontSize": 14}`
    indexes_to_remove: List[str], optional
        indexes to remove from the groupby selection.
    """

    alt.data_transformers.disable_max_rows()

    # Pre-processing
    predictor = probe.predictor.copy()
    predictor_metrics = probe._metrics.copy()
    estimates = fitted_model.estimates.copy()
    indexes = get_indexes_to_groupby(
        predictor_columns=predictor.columns,
        predictor_metrics=predictor_metrics,
        indexes_to_remove=indexes_to_remove,
    )
    predictor = predictor.merge(estimates, on=probe._index)
    predictor["normalized_date"] = month_diff(
        predictor["date"], predictor["t_0"]
    ).astype(int)
    for estimate in fitted_model._coefs + fitted_model._metrics:
        if pd.api.types.is_datetime64_any_dtype(predictor[estimate]):
            predictor[estimate] = predictor[estimate].dt.strftime("%Y-%m")
    predictor["normalized_c"] = predictor["c"].where(
        (predictor["normalized_date"] < 0) | (predictor["c_0"] == 0),
        predictor["c"] / predictor["c_0"],
    )
    predictor["model"] = 1
    predictor["model"] = predictor["model"].where(predictor["normalized_date"] >= 0, 0)
    predictor = filter_data(
        data=predictor,
        care_site_level=care_site_level,
        start_date=start_date,
        end_date=end_date,
        **kwargs,
    )
    if t_min:
        predictor = predictor[predictor.normalized_date >= t_min]
    if t_max:
        predictor = predictor[predictor.normalized_date <= t_max]

    # Get viz config
    probe_config = deepcopy(probe.get_viz_config("normalized_probe_plot"))
    model_config = deepcopy(
        fitted_model.get_viz_config("normalized_probe_plot", predictor=predictor)
    )
    if probe_line_config is None:
        probe_line_config = model_config["probe_line"]
    if model_line_config is None:
        model_line_config = model_config["model_line"]
    if error_line_config is None:
        error_line_config = model_config["error_line"]
    if estimates_selections is None:
        estimates_selections = model_config["estimates_selections"]
    if estimates_filters is None:
        estimates_filters = model_config["estimates_filters"]
    if main_chart_config is None:
        main_chart_config = probe_config["main_chart"]
    if chart_style is None:
        chart_style = probe_config["chart_style"]

    # Viz
    predictor["legend_model"] = (
        model_line_config.get("legend_title")
        if model_line_config.get("legend_title")
        else type(fitted_model).__name__
    )
    predictor["legend_predictor"] = probe_line_config["legend_title"]
    predictor["legend_error_band"] = error_line_config["legend_title"]
    index_selection, index_fields = create_groupby_selection(
        indexes=indexes,
        predictor=predictor,
    )
    base = alt.Chart(predictor)
    base = add_estimates_filters(
        base=base,
        estimates_filters=estimates_filters,
    )
    main_chart = generate_main_chart(
        base=base,
        main_chart_config=main_chart_config,
        index_selection=index_selection,
        index_fields=index_fields,
        x_axis_title=x_axis_title,
        y_axis_title=y_axis_title,
    )
    probe_line = generate_probe_line(
        main_chart=main_chart, probe_line_config=probe_line_config
    )
    error_line = generate_error_line(
        main_chart=main_chart, error_line_config=error_line_config
    )
    model_line = generate_model_line(
        main_chart=main_chart, model_line_config=model_line_config
    )
    main_chart = probe_line + error_line + model_line
    if index_selection:
        main_chart = main_chart.add_params(index_selection)

    for estimate_selection in estimates_selections:
        main_chart = main_chart.add_params(estimate_selection)

    main_chart = configure_style(chart=main_chart, chart_style=chart_style)

    if save_path:
        save_html(
            obj=main_chart,
            filename=save_path,
        )

    return main_chart



import altair as alt


def get_normalized_probe_dashboard_config(self, **kwargs):
    vertical_bar_charts, horizontal_bar_charts = self.generate_bar_chart_config(
        **kwargs
    )
    return dict(
        chart_style=chart_style,
        main_chart=normalized_main_chart,
        time_line=normalized_time_line,
        vertical_bar_charts=vertical_bar_charts,
        horizontal_bar_charts=horizontal_bar_charts,
    )


def get_estimates_densities_plot_config(self, **kwargs):
    vertical_bar_charts, horizontal_bar_charts = self.generate_bar_chart_config(
        **kwargs
    )
    return dict(
        chart_style=chart_style,
        vertical_bar_charts=vertical_bar_charts,
        horizontal_bar_charts=horizontal_bar_charts,
    )


def get_normalized_probe_plot_config(self, **kwargs):
    return dict(
        chart_style=chart_style,
        main_chart=normalized_main_chart,
    )


def get_probe_dashboard_config(self, **kwargs):
    vertical_bar_charts, horizontal_bar_charts = self.generate_bar_chart_config(
        **kwargs
    )
    return dict(
        chart_style=chart_style,
        main_chart=main_chart,
        time_line=time_line,
        vertical_bar_charts=vertical_bar_charts,
        horizontal_bar_charts=horizontal_bar_charts,
    )


def get_probe_plot_config(self, **kwargs):
    return dict(
        chart_style=chart_style,
        main_chart=main_chart,
    )


normalized_main_chart = dict(
    encode=dict(
        x=alt.X(
            "normalized_date:Q",
            title="Δt = (t - t₀) months",
            scale=alt.Scale(nice=False),
        ),
        y=alt.Y(
            "mean(normalized_c):Q",
            title="c(Δt) / c₀",
            axis=alt.Axis(grid=True),
        ),
        color=alt.Color(
            "value:N",
            title=None,
        ),
    ),
    properties=dict(
        height=300,
        width=900,
    ),
)

main_chart = dict(
    encode=dict(
        x=alt.X(
            "yearmonth(date):T",
            title="Time (Month Year)",
            axis=alt.Axis(tickCount="month", labelAngle=0, grid=True),
        ),
        y=alt.Y(
            "mean(c):Q",
            title="Completeness predictor c(t)",
            axis=alt.Axis(grid=True),
        ),
        color=alt.Color(
            "value:N",
            title=None,
        ),
        tooltip=[
            alt.Tooltip("value:N", title="Index"),
            alt.Tooltip("yearmonth(date):T", title="Date"),
            alt.Tooltip("mean(c):Q", title="c(t)", format=".2f"),
        ],
    ),
    properties=dict(
        height=300,
        width=900,
    ),
)

normalized_time_line = dict(
    encode=dict(
        x=alt.X(
            "normalized_date:Q",
            title="Δt = (t - t₀) months",
            scale=alt.Scale(nice=False),
        ),
        y=alt.Y(
            "mean(normalized_c):Q",
            title="c(Δt) / c₀",
        ),
    ),
    properties=dict(
        height=50,
        width=900,
    ),
)

time_line = dict(
    encode=dict(
        x=alt.X(
            "yearmonth(date):T",
            title="Time (Month Year)",
            axis=alt.Axis(tickCount="month", labelAngle=0, grid=True),
        ),
        y=alt.Y(
            "mean(c):Q",
            title="Completeness predictor c(t)",
        ),
    ),
    properties=dict(
        height=50,
        width=900,
    ),
)

error_line = dict(
    legend_title="Standard deviation",
    mark_errorband=dict(
        extent="stdev",
    ),
    encode=dict(
        stroke=alt.Stroke(
            "legend_error_band",
            title="Error band",
            legend=alt.Legend(
                symbolType="square",
                orient="top",
                labelFontSize=12,
                labelFontStyle="bold",
            ),
        ),
    ),
)

chart_style = dict(
    labelFontSize=12,
    titleFontSize=13,
)

# This is the basic viz configs if not overridden by the probe.
viz_configs = dict(
    normalized_probe_dashboard=get_normalized_probe_dashboard_config,
    probe_dashboard=get_probe_dashboard_config,
    estimates_densities_plot=get_estimates_densities_plot_config,
    normalized_probe_plot=get_normalized_probe_plot_config,
    probe_plot=get_probe_plot_config,
)



import catalogue

from edsteva.probes.condition.viz_configs import n_condition, per_condition, per_visit

normalized_probe_dashboard = catalogue.create(
    "edsteva.probes.condition.viz_configs", "normalized_probe_dashboard"
)
normalized_probe_dashboard.register(
    "per_visit_default", func=per_visit.get_normalized_probe_dashboard_config
)

normalized_probe_dashboard.register(
    "per_condition_default", func=per_condition.get_normalized_probe_dashboard_config
)

normalized_probe_dashboard.register(
    "n_condition", func=n_condition.get_normalized_probe_dashboard_config
)
probe_dashboard = catalogue.create(
    "edsteva.probes.condition.viz_configs", "probe_dashboard"
)
probe_dashboard.register("per_visit_default", func=per_visit.get_probe_dashboard_config)
probe_dashboard.register(
    "per_condition_default", func=per_condition.get_probe_dashboard_config
)
probe_dashboard.register("n_condition", func=n_condition.get_probe_dashboard_config)

estimates_densities_plot = catalogue.create(
    "edsteva.probes.condition.viz_configs", "estimates_densities_plot"
)
estimates_densities_plot.register(
    "per_visit_default", func=per_visit.get_estimates_densities_plot_config
)
estimates_densities_plot.register(
    "per_condition_default", func=per_condition.get_estimates_densities_plot_config
)
estimates_densities_plot.register(
    "n_condition", func=n_condition.get_estimates_densities_plot_config
)

normalized_probe_plot = catalogue.create(
    "edsteva.probes.condition.viz_configs", "normalized_probe_plot"
)
normalized_probe_plot.register(
    "per_visit_default", func=per_visit.get_normalized_probe_plot_config
)
normalized_probe_plot.register(
    "per_condition_default", func=per_condition.get_normalized_probe_plot_config
)
normalized_probe_plot.register(
    "n_condition", func=n_condition.get_normalized_probe_plot_config
)
probe_plot = catalogue.create("edsteva.probes.condition.viz_configs", "probe_plot")
probe_plot.register("per_visit_default", func=per_visit.get_probe_plot_config)
probe_plot.register("per_condition_default", func=per_condition.get_probe_plot_config)
probe_plot.register("n_condition", func=n_condition.get_probe_plot_config)

viz_configs = dict(
    normalized_probe_dashboard=normalized_probe_dashboard,
    probe_dashboard=probe_dashboard,
    estimates_densities_plot=estimates_densities_plot,
    normalized_probe_plot=normalized_probe_plot,
    probe_plot=probe_plot,
)



import catalogue

from .per_condition import compute_completeness_predictor_per_condition
from .per_visit import compute_completeness_predictor_per_visit

completeness_predictors = catalogue.create(
    "edsteva.probes.condition", "completeness_predictors"
)

completeness_predictors.register(
    "per_visit_default", func=compute_completeness_predictor_per_visit
)

completeness_predictors.register(
    "per_condition_default", func=compute_completeness_predictor_per_condition
)
